/*
        bzip2dll
                Copyright 1999 by Christian Ghisler (support@ghisler.com)
                This file is Freeware

        to be used with Windows Commander (www.ghisler.com)
*/


#define BZ_EXPORT           /* important: do not link to DLL! */
#include "bzlib.h"
#include <stdio.h>
#include <stdlib.h>

#ifdef _MSC_VER
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#endif

#ifdef _WIN32
#include <io.h>
#include <dos.h>
#endif

/* Error codes returned to calling application */
#define E_END_ARCHIVE     10            /* No more files in archive */
#define E_NO_MEMORY       11            /* Not enough memory */
#define E_BAD_DATA        12            /* Data is bad */
#define E_BAD_ARCHIVE     13            /* CRC error in archive data */
#define E_UNKNOWN_FORMAT  14            /* Archive format unknown */
#define E_EOPEN           15            /* Cannot open existing file */
#define E_ECREATE         16            /* Cannot create file */
#define E_ECLOSE          17            /* Error closing file */
#define E_EREAD           18            /* Error reading from file */
#define E_EWRITE          19            /* Error writing to file */
#define E_SMALL_BUF       20            /* Buffer too small */
#define E_EABORTED        21            /* Function aborted by user */
#define E_NO_FILES        22            /* No files found */
#define E_TOO_MANY_FILES  23            /* Too many files to pack */
#define E_NOT_SUPPORTED   24            /* Function not supported */

/* flags for unpacking */
#define PK_OM_LIST         0
#define PK_OM_EXTRACT      1

/* flags for ProcessFile */
#define PK_SKIP            0            /* Skip this file */
#define PK_TEST            1            /* Test file integrity */
#define PK_EXTRACT         2            /* Extract to disk */

/* Flags passed through ChangeVolProc */
#define PK_VOL_ASK         0            /* Ask user for location of next volume */
#define PK_VOL_NOTIFY      1            /* Notify app that next volume will be unpacked */

/* Flags for packing */

/* For PackFiles */
#define PK_PACK_MOVE_FILES 1    /* Delete original after packing        */
#define PK_PACK_SAVE_PATHS 2    /* Save path names of files             */

/* Returned by GetPackCaps */
#define PK_CAPS_NEW        1    /* Can create new archives              */
#define PK_CAPS_MODIFY     2    /* Can modify exisiting archives        */
#define PK_CAPS_MULTIPLE   4    /* Archive can contain multiple files   */
#define PK_CAPS_DELETE     8    /* Can delete files                     */
#define PK_CAPS_OPTIONS   16    /* Has options dialog                   */

typedef struct {
    char ArcName[260];
    char FileName[260];
    int Flags;
    int PackSize;
    int UnpSize;
    int HostOS;
    int FileCRC;
    int FileTime;
    int UnpVer;
    int Method;
    int FileAttr;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
  } tHeaderData;

typedef struct {
    char* ArcName;
    int OpenMode;
    int OpenResult;
    char* CmtBuf;
    int CmtBufSize;
    int CmtSize;
    int CmtState;
  } tOpenArchiveData;

/* Definition of callback functions called by the DLL
Ask to swap disk for multi-volume archive */
typedef int (__stdcall *tChangeVolProc)(char *ArcName,int Mode);
/* Notify that data is processed - used for progress dialog */
typedef int (__stdcall *tProcessDataProc)(unsigned char *Addr,int Size);

typedef struct {
  char ArchiveName[260];
  int ArchiveOpen;
  int ArchivePos;
  int ArchiveSize;
  int ArchiveDate;
  tChangeVolProc ChangeVolProc;
  tProcessDataProc ProcessDataProc;
} ArchiveRec;

int maxarchives=10;

/* Keep a list of currently open archives (up to a maximum of maxarchives) */
static ArchiveRec ArchiveList[11];   /* 0 for packing, 1-10 for unpacking! */
                                      /* Important: array must be initialized to all 0! */

int __stdcall OpenArchive(tOpenArchiveData* ArchiveData)
{
    FILE* ArchiveHandle;
    char sig[8];    /* File signature */
    int n;
    int i;
    int ArcHandle;
    int foundslot;
    unsigned short date, time;
#ifdef _MSC_VER
		struct _stat ftbuf;
		struct tm* fttm;
#endif		
 
    foundslot=0;
    for (i=1;i<=maxarchives;i++)
      if (ArchiveList[i].ArchiveOpen==0) {
        foundslot=1;
        ArcHandle=i;
        break;
      }

    if (foundslot==0) {
      ArchiveData->OpenResult=E_NO_MEMORY;
      return 0;
    }
		

    if((ArchiveHandle = fopen(ArchiveData->ArcName,"rb"))==NULL){
      ArchiveData->OpenResult=E_EOPEN;  
      return 0;
    }
    
    n = fread ( sig, 1, 8, ArchiveHandle );
    if (n != 8) {
      fclose(ArchiveHandle);
      ArchiveData->OpenResult=E_EREAD;
      return 0;
    }
    
    fseek(ArchiveHandle,0,SEEK_END);
    ArchiveList[ArcHandle].ArchiveSize=ftell(ArchiveHandle);
#ifdef _MSC_VER
		_stat(ArchiveData->ArcName,&ftbuf);
		fttm=localtime(&ftbuf.st_mtime);
		time=fttm->tm_hour<<11 | fttm->tm_min<<5 | fttm->tm_sec;
		date=(fttm->tm_year-80)<<9 | (fttm->tm_mon+1)<<5 | fttm->tm_mday;
#else
    _dos_getftime( ArchiveHandle->_handle, &date, &time );
#endif
		ArchiveList[ArcHandle].ArchiveDate=65536*date+time;

    fclose(ArchiveHandle);
    if (!((sig[0]=='B') && (sig[1]=='Z') && (sig[2]=='h'))) {
      ArchiveData->OpenResult=E_UNKNOWN_FORMAT;
      return 0;
    }
    strcpy(ArchiveList[ArcHandle].ArchiveName,ArchiveData->ArcName);
    ArchiveList[ArcHandle].ArchiveOpen=1;
    ArchiveList[ArcHandle].ArchivePos=0;
    return ArcHandle;  /* returns pseudohandle */
}

int __stdcall CloseArchive(int hArcData)
{
    if (hArcData>0 && hArcData<=maxarchives) {
      ArchiveList[hArcData].ArchiveOpen=0;
      return 0;
    } else  
      return E_ECLOSE;
}

int __stdcall ReadHeader(int hArcData,tHeaderData* HeaderData)
{
    if (hArcData>0 && hArcData<=maxarchives && ArchiveList[hArcData].ArchiveOpen) {
        char* p;
        char* p1;                  /* Build file name from archive name without extension */
        if (ArchiveList[hArcData].ArchivePos>0)
          return E_END_ARCHIVE; 
        p=strrchr(ArchiveList[hArcData].ArchiveName,'\\');
        if (p==NULL) p=ArchiveList[hArcData].ArchiveName; else p++;
        p1=strrchr(p,'.');
        if (p1 != NULL) {
          strncpy(HeaderData->FileName,p,p1-p);
        } else {
          strcpy(HeaderData->FileName,p);  
        }
        strcpy(HeaderData->ArcName,ArchiveList[hArcData].ArchiveName);
        HeaderData->PackSize=ArchiveList[hArcData].ArchiveSize;
        HeaderData->UnpSize=ArchiveList[hArcData].ArchiveSize;
        HeaderData->FileTime=ArchiveList[hArcData].ArchiveDate;
        ArchiveList[hArcData].ArchivePos+=1;  /* next file in multifile archive */
        return 0;
    } else {
        return E_EREAD;
    }    
}

typedef struct {
     FILE* handle;
} fptr;

typedef fptr* pfptr;

int __stdcall ProcessFile(int hArcData,int Operation,char* DestPath,char* DestName)
{
    if (hArcData<1 || hArcData>maxarchives)
        return E_EREAD;

    if (Operation==PK_SKIP)
        return 0;
        
    if (ArchiveList[hArcData].ArchiveOpen && ArchiveList[hArcData].ArchiveName[0] &&
        ((Operation==PK_EXTRACT && DestName !=NULL) || (Operation==PK_TEST))) {
        int len;
        char buff[0x1000];
        BZFILE *BZ2fp_r;
        FILE *fp_w;
        int aborted;
        int oldpos;
        int curpos;
        if (Operation==PK_EXTRACT)
                if ((fp_w = fopen(DestName,"wb"))==NULL)
                        return E_ECREATE;

        if ((BZ2fp_r = bzopen(ArchiveList[hArcData].ArchiveName,"rb"))==NULL)
           return E_EOPEN;

        aborted=0;
        oldpos=0;

        while((len=bzread(BZ2fp_r,buff,0x1000))>0 && aborted==0){
                if (Operation==PK_EXTRACT)
                        fwrite(buff,1,len,fp_w);
                if (ArchiveList[hArcData].ProcessDataProc) {
                        curpos=ftell(((pfptr)BZ2fp_r)->handle);
                        if (ArchiveList[hArcData].ProcessDataProc(NULL,curpos-oldpos)==0)
                        aborted=1;
                        oldpos=curpos;  
                }  
        }
        
        bzclose(BZ2fp_r);
        if (Operation==PK_EXTRACT)
                fclose(fp_w);
        if (aborted)
                return E_EABORTED;
        else if (len<0) {
                unlink(DestName);
                return E_BAD_ARCHIVE;
        } else
                return 0;  
    } else
        return E_EREAD;
}

void __stdcall SetChangeVolProc(int hArcData,tChangeVolProc pChangeVolProc)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ChangeVolProc=pChangeVolProc;
}

void __stdcall SetProcessDataProc(int hArcData,tProcessDataProc pProcessDataProc)
{
    int index=hArcData;
    if (hArcData<1 || hArcData>maxarchives)
      index=0;
    ArchiveList[index].ProcessDataProc=pProcessDataProc;    
}

int FileExists(char *FileName)
{
    FILE *fp_r;
    int res;
    fp_r = fopen(FileName,"rb");
    res=fp_r!=NULL;
    if (res) fclose(fp_r);
    return res;
}

int RenToTemp(char *FileName,char *RenamedName)    /* returns RenamedName */
{
    char *p;
    char ch1;
    char ch2;
    strcpy(RenamedName,FileName);
    p=RenamedName;
    while (p[0]) p++;
    strncat(RenamedName,".#00",259);
    p+=2;
    for (ch1='0';ch1<='9';ch1++) {
      p[0]=ch1;
      for (ch2='0';ch2<='9';ch2++) {
        p[1]=ch2;
        if (!rename(FileName,RenamedName))
          return 1;
      }
    }
    return 0;    /* rename failed! */
}

/* Functions for packing files */

int __stdcall PackFiles(char *PackedFile,char *SubPath,char *SrcPath,
      char *AddList,int Flags)
{
  char *p;
  int retval=0;

  p=AddList;
  if (!p || !p[0]) return E_NO_FILES;
  while (p[0]) p++;
  p++;
  if (p[0]) return E_TOO_MANY_FILES;
  /* Start to pack */
  {
        int len;
        int filethere;
        char buff[0x1000];
        char mode[10];
        char FullName[260];
        char OldName[260];
        BZFILE *BZ2fp_w;
        FILE *fp_r;
        int level = GetPrivateProfileInt("bzip2dll","CompressionRate",6,"wcplugin.ini");
        if (level<=0) level=1;
        if (level>=9) level=9;

        strcpy(FullName,SrcPath);
        strcat(FullName,AddList);

        filethere=FileExists(PackedFile);
        if (filethere) {
                /* Make sure the added filename is the same as the archive name without extension! */
                char *p;
                char TestName[260];
                strcpy(TestName,PackedFile);
                p=strrchr(TestName,'.');
                if (p) p[0]='\0';
                p=strrchr(TestName,'\\');
                if (p) p++;
                if (!p || stricmp(p,AddList)!=0)
                        return E_TOO_MANY_FILES;
                filethere=RenToTemp(PackedFile,OldName);
        }        

        if((fp_r = fopen(FullName,"rb"))==NULL) {
                if (filethere)
                        rename(OldName,PackedFile);
                return E_EOPEN;
        }        

        mode[0]='w';
        mode[1] = '0'+level;
        mode[2] = '\0';

        if ((BZ2fp_w = bzopen(PackedFile,mode))==NULL) {
                fclose(fp_r);
                if (filethere)
                        rename(OldName,PackedFile);
                return E_ECREATE;
        }
        while((len=fread(buff,1,0x1000,fp_r))>0 && retval==0) {
                if (ArchiveList[0].ProcessDataProc)
                        if (ArchiveList[0].ProcessDataProc((unsigned char*)AddList,len)==0) {
                                retval=E_EABORTED;
                                break;
                        }        
                if (bzwrite(BZ2fp_w,buff,len)==-1) {
                        retval=E_EWRITE;
                        break;
                }
        }
        bzclose(BZ2fp_w);
        fclose(fp_r);
        /* Error occurred -> delete new archive and restore old (if exists) */
        if (retval) {
                remove(PackedFile);
                if (filethere)
                        rename(OldName,PackedFile);
        } else {
                if (filethere)
                        unlink(OldName);
                /* Move to archive - only if packing was successful! */        
                if (Flags & PK_PACK_MOVE_FILES)
                        unlink(FullName);
        }
                
  }
  return retval; 
}

int __stdcall DeleteFiles(char *PackedFile,char *DeleteList)
{
  /* Since this archive contains only one file anyway, just delete the whole archive! */
  if (DeleteList && DeleteList[0])
    if (unlink(PackedFile))
      return E_EWRITE;
    else
      return 0;
  return E_BAD_DATA;
}

int __stdcall GetPackerCaps()
{
  return PK_CAPS_NEW | PK_CAPS_MODIFY | PK_CAPS_DELETE | PK_CAPS_OPTIONS;
}

HANDLE hBrush;

BOOL CALLBACK DialogProc(HWND hwndDlg,UINT  uMsg,WPARAM wParam,LPARAM  lParam)
{
   int i;
   char num[4];
   char ch;
   num[1]='\0';
   switch (uMsg) {
     case WM_INITDIALOG:
       for (ch='1';ch<='9';ch++) {
           num[0]=ch;
           SendDlgItemMessage(hwndDlg,103,CB_ADDSTRING,0,(LPARAM)&num);
       }
       i=GetPrivateProfileInt("bzip2dll","CompressionRate",6,"wcplugin.ini");
       SendDlgItemMessage(hwndDlg,103,CB_SETCURSEL,i-1,0);
       hBrush=CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
       return TRUE;
       break;
     case WM_CTLCOLORDLG:
     case WM_CTLCOLORSTATIC:
     case WM_CTLCOLORBTN:
       SetBkColor((HDC)wParam,GetSysColor(COLOR_BTNFACE));
       SetTextColor((HDC)wParam,GetSysColor(COLOR_BTNTEXT));
       return (BOOL)hBrush;
       break;
     case WM_COMMAND:
       switch (wParam) {
         case 104:    /* OK  */
           num[0]=(char)('1'+SendDlgItemMessage(hwndDlg,103,CB_GETCURSEL,0,0));
           WritePrivateProfileString("bzip2dll","CompressionRate",num,"wcplugin.ini");
           DeleteObject(hBrush);
           EndDialog(hwndDlg,1);
         case 105:    /* Cancel */
           DeleteObject(hBrush);
           EndDialog(hwndDlg,0);
       }
     break;
   }
   return FALSE;
}



void __stdcall ConfigurePacker(HANDLE ParentHandle, HANDLE hinstance)
{	

   /* You may ask for packing (NOT unpacking!) parameters here */
   DialogBox(hinstance,"CONFDLG",ParentHandle,(DLGPROC)&DialogProc);
}

